/*

  read image: NR rows, NC columns of floats. (values usually in [0,1]).
  for each pixel: 
    define neighbors
    compute similarity between the pixel and its neighbors (use DSCALE, ISCALE)
    scale and round weights (similarities) using SCALING_FACTOR.
    print sparse matrix w(i,j)

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define SCALING_FACTOR 1e+4
#define MIN_SIMILARITY 100
#define ONLY4
#define RAND_FRIENDS 4
#define DRANGE 12
#define DSCALE 64.0
#define ISCALE 0.01
#define SQR(x) ((x)*(x))

struct pix {
  int x;
  int y;
  int rnd;
};

struct friend {
  int name;
  int sim;
  struct friend *next;
};

int read_image(char *fname);
int init_circle();
int neighbors(int i, int j);
void insert(int n, int m, int isim);
int in_list(int n, int m);
double similarity(int i1, int j1, int i2, int j2);
int pix_order(const void *x, const void *y);
void error(char *message);

int NR, NC;
int NUM_CIRCLE2;
float **Image;
#ifdef ONLY4
 struct pix circle1[4], circle2[SQR(2*DRANGE+1)-4];
#else
 struct pix circle1[8], circle2[SQR(2*DRANGE+1)-8];
#endif
struct friend** sparse;

main(int argc, char *argv[])
{
  char *fname;
  int i,j;
  int npix;
  int count;
  struct friend *p;


  /*  srand(98987764); */

  /* command line */
  if (argc != 2) {
    fprintf(stderr,"Usage: %s image-file(with header)\n", argv[0]);
    exit(1);
  }
  fname = argv[1];

  /* allocate and read Image, init other global vartiables */
  npix = read_image(fname);
  NUM_CIRCLE2 = init_circle();
  sparse = (struct friend **)malloc(npix*sizeof(struct friend *));
  if (!sparse) error("allocation error");
  for (i=0; i<npix; i++)
    sparse[i] = NULL;
  

  /* build sparse similarity matrix */
  count=0;
  for (i=0; i<NR; i++) {
    for (j=0; j<NC; j++) {
      count += neighbors(i,j);
    }
  }

  /* print sparse matrix */
  printf("# %d %d\n", npix, count);
  for (i=0; i<npix; i++) {
    p = sparse[i];
    while (p!=NULL) {
      printf("%d %d %d\n", i, p->name, p->sim);
      p = p->next;
    }
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double similarity(int i1, int j1, int i2, int j2)
{
  double d_square, dI_square, dI;
  double sim1, sim2;

  d_square = (double)(SQR(i1-i2) + SQR(j1-j2));
  sim1 = exp(-d_square/DSCALE);

  dI = (double)(Image[i1][j1] - Image[i2][j2]);
  dI_square = (double)SQR(dI);
  sim2 = exp(-dI_square/ISCALE);

  return(sim1*sim2);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int neighbors(int i, int j)
{
  int count_8, count_rand;
  int i2, j2, k;
  int n, m;
  int isim;
  double sim;

  count_8 = count_rand = 0;
  n = i*NC+j;

  /* try to select neighbors from circle1 */
#ifdef ONLY4
  for (k=0; k<4; k++) {
#else
  for (k=0; k<8; k++) {
#endif
    i2 = i + circle1[k].x;
    j2 = j + circle1[k].y;
    if (i2>=0 && i2<NR && j2>=0 && j2<NC) {
      m = i2*NC+j2;

      if (!in_list(n,m)) { /* check if n=(i,j) in the list m=(i2,j2) */
	sim = similarity(i,j,i2,j2);
	isim = (int)(SCALING_FACTOR * sim);
	if (isim > MIN_SIMILARITY) {
	  insert(n,m,isim); /* insert m (+isim) to the list n */
	  count_8++;
	}
      }
    }
  }

  if (RAND_FRIENDS==0)
    return(count_8);

  /* scramble neighbor list in circle2 */
  for (k=0; k<NUM_CIRCLE2; k++)
    circle2[k].rnd = rand();
  qsort(circle2, NUM_CIRCLE2, sizeof(struct pix), pix_order);

  /* try to select neighbors from circle2 */
  for (k=0; k<NUM_CIRCLE2; k++) {
    i2 = i + circle2[k].x;
    j2 = j + circle2[k].y;
    if (i2>=0 && i2<NR && j2>=0 && j2<NC) {
      m = i2*NC+j2;

      if (!in_list(n,m)) { /* check if n=(i,j) in the list m=(i2,j2) */
	sim = similarity(i,j,i2,j2);
	isim = (int)(SCALING_FACTOR * sim);
	if (isim > MIN_SIMILARITY) {
	  insert(n,m,isim); /* insert m (+isim) to the list of n */
	  count_rand++;
	}
      }
    }
    if (count_rand==RAND_FRIENDS)
      break;
  }

  /* return the number of neighbors added for pixel (i,j) */
  return(count_rand + count_8);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* insert m (+isim) to the list of n */
void insert(int n, int m, int isim)
{
  struct friend *new;

  new = (struct friend *) malloc(sizeof(struct friend));
  if (!new) error("allocation error");

  new->name = m;
  new->sim  = isim;
  new->next = sparse[n];
  sparse[n] = new;

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* check if n=(i,j) in the list m=(i2,j2) */
int in_list(int n, int m)
{
  struct friend *p;

  p = sparse[m];
  while (p!=NULL) {
    if (p->name == n)
      return(1);
    p = p->next;
  }

  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int init_circle()
{
  int rsqr, rsqr_max;
  int count1, count2;
  int i,j;

  count1=count2=0;
  rsqr_max = DRANGE * DRANGE;
  for (i= -DRANGE; i<=DRANGE; i++) {
    for (j= -DRANGE; j<=DRANGE; j++) {
      if (!(i==0 && j==0)) {
	rsqr = SQR(i) + SQR(j);
#ifdef ONLY4
	if (rsqr < 1.1) { /* the 4-neighborhood */
#else
	if (rsqr < 2.1) { /* the 8-neighborhood */
#endif
	  circle1[count1].x = i;
	  circle1[count1].y = j;
	  count1++;
	} else if (rsqr <= rsqr_max) {
	  circle2[count2].x = i;
	  circle2[count2].y = j;
	  count2++;
	}
      }
    }
  }

  return(count2);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int read_image(char *fname)
{
  FILE *pf;
  int i,j;

  if ((pf = fopen(fname,"r")) == NULL)
    error("cannot open file");
  fscanf(pf,"%*s %d %d",&NR, &NC);
  fprintf(stderr,"found image size: %d X %d\n",NR,NC); 

  /* allocate Image */
  Image = (float **)malloc(NR*sizeof(float *));
  if (!Image) error("allocation error");
  for (i=0; i<NR; i++) {
    Image[i] = (float *)malloc(NC*sizeof(float));  
    if (!Image[i]) error("allocation error");
  }
  

  /* read points */
  for (i=0; i<NR; i++)
    for (j=0; j<NC; j++)
      fscanf(pf,"%f",&Image[i][j]);
  fclose(pf);

  return(NR*NC);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void error(char *message)
{
  fprintf(stderr,"Quit due to: %s\n", message);
  exit(1);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int pix_order(const void *x, const void *y)
{
  struct pix *a, *b;

  a = (struct pix *)x;
  b = (struct pix *)y;

  /* sort by increasing order */
  if (a->rnd < b->rnd) return (-1);
  if (a->rnd > b->rnd) return ( 1);
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
