/*

  read vectorial data.
  compute Euclidean distances d(i,j).
  transform d(i,j) -> w(i,j) by Gaussian decay, use KNN as a parameter.
  eliminate all weights w(i,j) which are below a threshold theta.
  scale and round weights by SCALING_FACTOR.
  print sparse matrix w(i,j) into stdout.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define SQR(x) ((x)*(x))
#define SCALING_FACTOR  1.0e+4

struct scramble {
  int u;
  int v;
  float weight;
  int order;
};


float find_dknn(float *v, int n);
int get_dist(float ***dist, char *fname);
void error(char *message);
int scrmbl_order(const void *x, const void *y);

int KNN;

main(int argc, char *argv[])
{
  char *fname;
  float theta;
  int np, ne;
  int i, j;
  int isim;
  float ls, ls2;
  float **dist;
  struct scramble *wksp;

  /* command line */
  if (argc != 4) {
    fprintf(stderr,"Usage: %s <vdat> <KNN> <theta>\n", argv[0]);
    exit(1);
  }
  fname = argv[1];
  KNN = atoi(argv[2]);
  theta = atof(argv[3]);
  if (1./theta > SCALING_FACTOR)
    error("SCALING_FACTOR is too small for the requested threshold");


  /* read vectorial data, allocate and compute Euclidean dist matrix */
  np = get_dist(&dist, fname);


  /* transform Euclidean distances to weights in the [0,1] interval */

  /* 1. compute average distance to the K-nearest-neighbor */
  /*    find_dknn() assumes that diagonal terms in dist are zero */
  ls = 0.0;
  for (i=0; i<np; i++) {
    ls += find_dknn(dist[i], np);
  }
  ls /= (float)np;
  ls2 = SQR(ls);
  /*------------------------------------------------*/
  fprintf(stderr,"local scale (ls) = %f\n",ls);
  /*------------------------------------------------*/
 

  /* 2. replace distances with weights ONLY IN THE LOWER TRIANGLE */
  for (i=0; i<np; i++) {
    for (j=0; j<i; j++) {
      dist[i][j] = exp(-SQR(dist[i][j])/ls2);
    }
  }


  /* alloc workspace to scramble edges */
  ne = (np*(np-1))>>1;
  wksp = (struct scramble *)malloc(ne*sizeof(struct scramble)); 
  if (!wksp) error("allocation error");

  /* count how many edges are above threshold, and store them in wksp */
  ne = 0;
  for (i=0; i<np; i++) {
    for (j=0; j<i; j++) {
      if (dist[i][j] > theta) {
	wksp[ne].u = i; 
	wksp[ne].v = j;
	wksp[ne].weight = dist[i][j];
	wksp[ne].order = rand();
	ne++;
      }
    }
  }


  /* scramble edges and print scaled (integer) weights */
  printf("# %d %d\n", np, ne);
  qsort(wksp, ne, sizeof(struct scramble), scrmbl_order);
  for (i=0; i<ne; i++) {
    isim = (int)(SCALING_FACTOR * wksp[i].weight);
    
    if (SCALING_FACTOR * wksp[i].weight - isim > 0.5)
      isim++;
    
    printf("%d %d %d\n", wksp[i].u, wksp[i].v, isim);
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int get_dist(float ***dist, char *fname)
{
  FILE *pf;
  int nvec, dim;
  int i,j,k;
  float **Vec;
  float dk, d;

  if ((pf = fopen(fname,"r")) == NULL)
    error("cannot open file");
  fscanf(pf,"%*s %d %d", &nvec, &dim);

  Vec = (float **)malloc(nvec*sizeof(float *));
  if (!Vec) {fprintf(stderr,"allocation error\n"); exit(1);}
  for (i=0; i<nvec; i++) {
    Vec[i] = (float *)malloc(dim*sizeof(float));  
    if (!Vec[i]) {fprintf(stderr,"allocation error\n"); exit(1);}
  }

  *dist = (float **)malloc(nvec*sizeof(float *));
  if (!(*dist)) {fprintf(stderr,"allocation error\n"); exit(1);}
  for (i=0; i<nvec; i++) {
    (*dist)[i] = (float *)malloc(nvec*sizeof(float));
    if (!(*dist)[i]) {fprintf(stderr,"allocation error\n"); exit(1);}
  }

  /* read points */
  for (i=0; i<nvec; i++)
    for (j=0; j<dim; j++)
      fscanf(pf,"%f",&Vec[i][j]);
  fclose(pf);


  /* compute distances */
  for (i=0; i<nvec; i++) {
    (*dist)[i][i] = 0.0;
    for (j=0; j<i; j++) {
      d = 0.0;
      for (k=0; k<dim; k++) {
	dk = Vec[i][k] - Vec[j][k];
	d += SQR(dk);
      }
      d = sqrt(d);
      
      (*dist)[i][j] = (*dist)[j][i] = d;
    }
  }


  /* free Vec */
  for (i=nvec-1; i>=0; i--) 
    free(Vec[i]);
  free(Vec);

  return(nvec);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
float find_dknn(float *v, int n)
{
  int *small; /* add 1 to KNN since v includes the zero self-distance */
  int i, j;
  int top; /* the index in "small" which refers to a largest value */
  float k;

  small = (int *)malloc((KNN+1)*sizeof(int));
  top=0;
  for (i=0; i<KNN+1; i++) {
    small[i] = i;
    if (v[i] > v[small[top]]) {
      top = i;
    }
  }

  for (i=KNN+1; i<n; i++) {
    if (v[i] < v[small[top]]) {
      small[top] = i;
      top=0;
      for (j=1; j<KNN+1; j++)
	if (v[small[j]] > v[small[top]]) 
	  top=j;
    }
  }

  k = v[small[top]];
  free(small);
  return(k);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int scrmbl_order(const void *x, const void *y)
{
  struct scramble *a, *b;

  a = (struct scramble *)x;
  b = (struct scramble *)y;

  /* sort by increasing order */
  if (a->order < b->order) return (-1);
  if (a->order > b->order) return ( 1);
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void error(char *message)
{
  fprintf(stderr,"Quit due to: %s\n", message);
  exit(1);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
