#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ----------------------- types ------------------------------------ */

struct params {     /* command line parameters */
  char  *fsim;      /* file name: sparse weighted graph (sdat) */
  char  *fmed;      /* file name: medians */ 
  char  *fpeak;     /* file name: peaks */
  char  *fdata;     /* file name: vectorial data or image header */
  int   graph;      /* 1=make graph, 0=report partitions */
  int   correct0;   /* 1=relabel zeros (default) */
  int   printype;   /* 0=indices, 1=vectors, 2=image */
  int   min_size;   /* label 0 is given for clusters below this size */
  float min_height; /* min height for stored partitions */
  int   report_by;  /* 0=all, 1=level(r), 2=all n largest, 3=the n-th largest*/
  int   report_val; /* value of r or n (see 'report_by' above) */
  int   quite;      /* 1=don't ask for output filenames, print to stdout */
};

struct edge {
  int u;
  int v;
  int weight;
};

struct xrs {
  int leaf;
  struct xrs *next;
};

struct vcross {
  int vn;
  struct xrs *leaves;
  struct xrs *tail;
  struct vcross *next;
};

struct peaklist {
  int r;
  float height;
  struct peaklist *next;
};

struct partition {
  float height;
  int level;
  int report;
  int *Labels;
  struct partition *prev;
  struct partition *next;
};

/* ----------------------- functions ---------------------------------*/
/* connected components and the T(r) measure */
void init_global(char *fsim, char *fmed);
struct peaklist *init_peaks(int graph, char *fpeak);
struct xrs *find_crossing(int u1, int u2);
/* zero correction */
void top_down(struct partition *clust, int new_grain_ok);
int  expand(struct vcross *p, int new_name, int *labels);
int  de_expand(struct vcross *p, int bad_name, int *labels);
void correct_zeros(int *refs, int *labels, int new_grain_ok);
int  conflict(int u, int v, int n, int *labels, int *refs);
/* storing and printing partitions */
void store_partition(struct partition *clust, int level, float deltaTr);
void select_levels(struct partition * clust, int by, int val);
void print_indices(FILE *pf, int *Labels);
void print_vectors(FILE *pf, int *Labels, char *fdata, int shift);
void print_image(FILE *pf, int *Labels, char *fdata, int printype, int shift);
/* auxiliary functions */
void command_line(int argc, char *argv[], struct params *Par);
int order_vcross(const void *x, const void *y);
int order_partition(const void *x, const void *y);
int leaves_order(const void *x, const void *y);
void error(char *message);

/* ----------------------- global variables --------------------------*/
struct edge *leaves;
struct vcross **Urep;
struct vcross  *Vrep;
struct xrs *UVleaves;
int *numv;
int *Pname;
int *Psize;
int total_vcross;
int NP; /* number of nodes */
int NE; /* number of edges */
int MIN_CLUSTER_SIZE;
float num_ecross, init_ecross, delta_ecross, delta_ecross2;


main(int argc, char *argv[])
{
  int r, r0;
  int u,v;
  int leaf;
  int stopcase;
  float  deltaTr;
  struct peaklist *peaks, *curpeak, *q;
  struct edge *p;
  struct xrs *xedge;
  struct partition *clust, *inclust;
  struct params Par;
  FILE *pfout;
  char fname[24];

  /* command line */
  command_line(argc, argv, &Par);  
  MIN_CLUSTER_SIZE = Par.min_size;
  init_global(Par.fsim, Par.fmed);
  peaks = curpeak = init_peaks(Par.graph, Par.fpeak);

  /* allocate top of partition list (if no -g option used) */
  if (!Par.graph) {
    clust = (struct partition *)malloc(sizeof(struct partition));
    if (!clust) error("allocation error (clust)");
    clust->prev = clust->next = NULL;
    clust->Labels = NULL;
  }

  /* find connected components for all r values */
  /* store either partitions list ("clust") or peaks graph */
  p = leaves+NE-1;
  r = p->weight;
  while (r >= 0 || r == -1) {

    r0 = r;
    deltaTr = 0.0;
    while (r==r0 || r == -1) {
      if (r != -1) {
	u = Pname[p->u];
	while (u != Pname[u])
	  u = Pname[u];
	
	v = Pname[p->v];
	while (v != Pname[v])
	  v = Pname[v];
	
	xedge = find_crossing(u,v);
	while (xedge != NULL) {
	  leaf = xedge->leaf;
	  leaves[leaf].weight = -1;
	  xedge = xedge->next;
	}

	deltaTr += delta_ecross2/init_ecross;
      }
      stopcase =  (--p < leaves || total_vcross==0);
      if (stopcase)
	break;
      r = p->weight;
    }

    if (Par.graph && deltaTr>0.0) { /* deltaTr=0 possible with delta_ecross2*/
      q = (struct peaklist *)malloc(sizeof(struct peaklist));
      if (!q) error("allocation error (peaklist)");
      if (curpeak == NULL)
	peaks = q;
      else
	curpeak->next = q;
      q->r = r0;
      q->height = deltaTr;
      q->next = NULL;
      curpeak = q;
    }


    if (!Par.graph) {
      if (Par.report_by==1) { /* a specific r partition is wanted */
	if (Par.correct0) {
	  /* need smaller r's (at peaks) for zero correction */
	  if (r == curpeak->r ) {
	    if (curpeak->height > Par.min_height && r<=Par.report_val)
	      store_partition(clust,r,curpeak->height);
	    curpeak = curpeak->next;
	    if (curpeak == NULL)
	      stopcase = 1;
	  } else {
	    if (r==Par.report_val)
	      store_partition(clust,r,deltaTr);
	  }
	} else { 
	  /* store only one partition */
	  if (r<=Par.report_val) {
	    /* store_partition(clust,r,-1.0); /* height -1.0 is dummy */
	    store_partition(clust,r,deltaTr); 
	    stopcase = 1;
	  }
	}
      } else {
	if (r == curpeak->r ) { 
	  /* check r (not r0) ==> store for r just above r0=r(peak) */
	  if (curpeak->height > Par.min_height)
	    store_partition(clust,r,curpeak->height);
	  curpeak = curpeak->next;
	  if (curpeak == NULL)
	    stopcase = 1;
	}
      }
    }
       


    if (stopcase)
      break;
  }

  
  /* print peaks graph */
  if (Par.graph) {
    curpeak = peaks;
    while (curpeak != NULL) {
      printf("%d %f\n", curpeak->r, curpeak->height);
      curpeak = curpeak->next;
    }
    exit(0);
  } 

  /* relabel small clusters */
  if (Par.correct0)
    top_down(clust, (Par.correct0==2));

  /* print output in desired format */
  if (clust->Labels == NULL)
    if (Par.report_by==1) { /* patch: the desierd r was marked -1 and missed */
      store_partition(clust,r,-1.0); /* height -1.0 is dummy */
    } else
      error("no partitions found");


  select_levels(clust, Par.report_by, Par.report_val);

  inclust = clust;
  while (inclust != NULL) {
    if (inclust->report) {

      if (Par.quite) {
	printf("\n****************************************************\n");
	pfout = stdout;
      } else {
	if (inclust->height >= 0.0)
	  fprintf(stderr,"\npartition:  r = %d,   peak height = %f\n", 
		  inclust->level, inclust->height);
	else
	  fprintf(stderr,"\npartition:  r = %d,   height below threshold\n", 
		  inclust->level);
	fprintf(stderr,"enter file name to store partition\n");
	fprintf(stderr,"(or 0 to append to stdout): ");
	fscanf(stdin,"%s", fname);
	fprintf(stderr,"\n");
	if (strcmp(fname,"0")==0) {
	  printf("\n****************************************************\n");
	  pfout = stdout;
	} else {
	  if ((pfout=fopen(fname,"w"))==NULL) 
	    error("cannot open file for writing");
	}
      }

      switch (Par.printype) {
      case 0:
	print_indices(pfout, inclust->Labels);
	break;
      case 1:
	print_vectors(pfout, inclust->Labels, Par.fdata, (Par.correct0==2));
	break;
      case 2:
	print_image(pfout, inclust->Labels, Par.fdata, 2, (Par.correct0==2));
	break;
      case 3:
	print_image(pfout, inclust->Labels, Par.fdata, 3, (Par.correct0==2));
	break;
      default:
	error("no such output format");
	break;
      }
    }
    inclust = inclust->next;
  }  

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void select_levels(struct partition * clust, int by, int val)
{
  struct partition *p, *wksp, *q;
  int found, num, i;

  if (by == 0) { /* report all */
    p = clust;
    while (p != NULL) {
      p->report = 1;
      p = p->next;
    }  
    return;
  }

  if (by == 1) { /* report by level r */
    found = 0;
    p = clust;
    while (p != NULL) {
      if (p->level == val)
	p->report = found = 1;
      else
	p->report = 0;
      p = p->next;
    }

    if (!found) {
      q = NULL;
      p = clust; 
      while (p != NULL) {
	if (p->level < val)
	  break;
	q = p;
	p = p->next;
      }
      if (p==NULL)
	q->report = 1;
      else
	p->report = 1;
    }
    return;    
  }

  /* the rest of the code here is for by=2,3 */

  /* count number of partitions found */
  num=0;
  p = clust;
  while (p != NULL) {
    num++;
    p = p->next;
  }
  if (by==3 && val>num)
    error("option -o asks for a peak that does not exist");

  /* allocate workspace and sort partitions by height */
  wksp = (struct partition *)malloc(num*sizeof(struct partition));
  if (!wksp) error("allocation error (wksp in select_levels)");
  num=0;
  p = clust;
  while (p != NULL) {
    wksp[num].height = p->height;
    wksp[num].next = p;
    num++;
    p = p->next;
  }  
  qsort(wksp, num, sizeof(struct partition), order_partition);    

  /* scan partitions by height order */
  for (i=0; i<num; i++) {
    if ((by == 2 && i < val) || (by == 3 && i == val-1))
      (wksp[i].next)->report = 1;
    else
      (wksp[i].next)->report = 0;
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void top_down(struct partition *clust, int new_grain_ok)
{
  struct partition *top, *down;
  int *together;
  int i;


  /* find top level */
  top = clust;
  while (top->next != NULL)
    top = top->next;  

  /* correct top level using a dummy partition to a single cluster */
  together = (int *)malloc(NP*sizeof(int));
  if (!together) error("allocation error (labels)");
  for (i=0; i<NP; i++)
    together[i] = 1;
  correct_zeros(together, top->Labels, new_grain_ok);
  free(together);

  down = top->prev;
  while (down != NULL) {
    /* correct the zero labels in "down" according to "top" */
    correct_zeros(top->Labels, down->Labels, new_grain_ok);

    top = down;
    down = top->prev;
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int expand(struct vcross *p, int new_name, int *labels)
{
  struct vcross *q;
  int count;

  count = 0;
  while (p != NULL) {
    if (labels[p->vn] == 0) {
      labels[p->vn] = new_name;
      count++;
      if ((q=Urep[p->vn]) != NULL)
	count += expand(q, new_name, labels);
    }
    p = p->next;
  }

  return(count);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int de_expand(struct vcross *p, int bad_name, int *labels)
{
  struct vcross *q;
  int count;

  count = 0;
  while (p != NULL) {
    if (labels[p->vn] == bad_name) {
      labels[p->vn] = 0;
      count++;
      if ((q=Urep[p->vn]) != NULL)
	count += de_expand(q, bad_name, labels);
    }
    p = p->next;
  }

  return(count);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void correct_zeros(int *refs, int *labels, int new_grain_ok)
{
  int u,v,n;
  int count;
  int bg;
  struct vcross *p;
  int tag;

  for (n=0; n<NP; n++) {
    Urep[n] = NULL;
  }

  count= -1;
  for (n=NE-1; n>=0; n--) {
    u = leaves[n].u;
    v = leaves[n].v;
    if ((labels[u]==0 || labels[v]==0) && refs[u]==refs[v]) {

      if (labels[u]==0 && labels[v]!=0) {
	bg = conflict(u,v,n,labels,refs);
	if (!bg) { 
	  labels[u] = labels[v];
	  expand(Urep[u], labels[v], labels);
	}
      }
      
      else if (labels[v]==0 && labels[u]!=0) {
	bg = conflict(v,u,n,labels,refs);
	if (!bg) { 
	  labels[v] = labels[u];
	  expand(Urep[v], labels[u], labels);
	}
      }
      
      else if (labels[v]==0 && labels[u]==0) {
	if (++count == NE-1)
	  error("no more place in Vrep to store new edges");
	p = Vrep+count;
	p->vn = v;
	p->next = Urep[u];
	Urep[u] = p;      
	
	if (++count == NE-1)
	  error("no more place in Vrep to store new edges");
	p = Vrep+count;
	p->vn = u;
	p->next = Urep[v];
	Urep[v] = p;      
      }
    }
  }


  /* clusters in the background */
  if (new_grain_ok) {
    tag = NP+1;
    for (n=0; n<NP; n++) {
      if (labels[n]==0) {
	labels[n] = tag;
	count = 1;
	count += expand(Urep[n], tag, labels);
	if (count < MIN_CLUSTER_SIZE) { /* back to bg... */
	  labels[n] = 0;
	  de_expand(Urep[n], tag, labels); 
	} else
	  tag++;
      }
    }
  }



  /* points in conflict */
  for (n=0; n<NP; n++)
    if (labels[n]==NP)
      labels[n]=0;

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* for each i connected to u (including u itself), check:
   - that no label other than labels[v] is suggested by the same r, and
   - that i was not previously in conflict.
   return 1 in case of conflict.
   In this case mark labels[i]=NP for all those i's */
int conflict(int u, int v, int n, int *labels, int *refs)
{
  int bg,i,r,u1,v1,m;
  struct vcross *p;

  bg=0;
  p=Urep[u];
  while (p != NULL && !bg) {
    i = p->vn;
    if (labels[i] == NP)
      bg=1;

    m=n;
    r = leaves[m--].weight;
    while (r == leaves[m].weight && m>=0 && !bg) {
      u1 = leaves[m].u;
      v1 = leaves[m].v;

      if (u1==i && refs[v1]==refs[i] && labels[v1]>0) {
	if (labels[v] != labels[v1])
	  bg = 1;
      } else if (v1==i && refs[u1]==refs[i] && labels[u1]>0) {
	if (labels[v] != labels[u1])
	  bg = 1;
      }

      m--;
    }
    
    p = p->next;
  }  

  if (bg) {
    while (p != NULL) {
      labels[p->vn]=NP;
      p = p->next;
    }
  }

  return(bg);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void store_partition(struct partition *clust, int level, float deltaTr)
{
  struct partition *p;
  int i;
  int tag, cluster, num_big;

  /* find last item */
  p = clust;
  while (p->next != NULL)
    p = p->next;

  /* extend partition list by one */
  if (p->Labels != NULL) {
    p->next = (struct partition *)malloc(sizeof(struct partition));
    if (!p->next) error("allocation error");
    p->next->prev = p;
    p->next->next = NULL;
    p = p->next;
  }

  /* alloc space for new labels and initialize */
  p->Labels = (int *)malloc(NP*sizeof(int));
  if (!(p->Labels)) error("allocation error (labels)");
  for (i=0; i<NP; i++)
    (p->Labels)[i] = -1;

  /* find labels */
  num_big = 0;
  cluster = 0;
  for (i=0; i<NP; i++) {
    tag = Pname[i];
    while (tag != Pname[tag])
      tag = Pname[tag];

    if ((p->Labels)[tag] == -1) {
      if (Psize[tag] < MIN_CLUSTER_SIZE)
	(p->Labels)[tag] = 0;           /* assign "background label" */
      else {
	(p->Labels)[tag] = (++cluster); /* assign new label */
	num_big++;
      }
    }
    (p->Labels)[i] = (p->Labels)[tag];
  }
  /*
  fprintf(stderr,"number of clusters with more than %d points:  %d\n",
	  MIN_CLUSTER_SIZE, num_big);
	  */

  /* store level and peak height */
  p->height = deltaTr;
  p->level = level;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void error(char *message)
{
  fprintf(stderr,"Quit due to: %s\n", message);
  exit(1);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
struct peaklist *init_peaks(int graph, char *fpeak)
{
  FILE *pf;
  int r;
  float h;
  struct peaklist *p, *last, *top;

  if (graph) {
    return(NULL);
  }

  /* read peaks from file */
  if ((pf=fopen(fpeak,"r"))==NULL)
    error("cannot open peaks file");

  last = NULL;
  while (fscanf(pf, "%d %f", &r, &h) != EOF) {
    p = (struct peaklist *)malloc(sizeof(struct peaklist));
    if (!p) error("allocation error (peaklist)");
    if (last == NULL)
      top = p;
    else
      last->next = p;
    p->r = r;
    p->height = h;
    p->next = NULL;
    last = p;
  }

  fclose(pf);
  return(top);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void init_global(char *fsim, char *fmed)
{
  FILE *pfsim, *pfmed;
  int i;
  void init_reps2l();

  if ((pfmed=fopen(fmed,"r"))==NULL)
    error("cannot open medians file");
  if ((pfsim=fopen(fsim,"r"))==NULL)
    error("cannot open file (sparse similarity matrix)");
  fscanf(pfsim,"%*s %d %d", &NP, &NE);

  /* memory allocations */
  leaves = (struct edge *)malloc(NE*sizeof(struct edge));
  if (!leaves) error("allocation error");
  Urep = (struct vcross **)malloc(NP*sizeof(struct vcross *));
  if (!Urep) error("allocation error");
  Vrep = (struct vcross *)malloc(NE*sizeof(struct vcross));
  if (!Vrep) error("allocation error");
  UVleaves = (struct xrs *)malloc(NE*sizeof(struct xrs));
  if (!UVleaves) error("allocation error");
  numv = (int *)malloc(NP*sizeof(int));
  if (!numv) error("allocation error");
  Pname = (int *)malloc(NP*sizeof(int));
  if (!Pname) error("allocation error");
  Psize = (int *)malloc(NP*sizeof(int));
  if (!Psize) error("allocation error");

  /* initalize leaves : read from files and sort by weight */
  for (i=0; i<NE; i++) {
    fscanf(pfsim,"%d %d %*d", &(leaves[i].u), &(leaves[i].v));
    fscanf(pfmed,"%d", &(leaves[i].weight));
  }
  fclose(pfsim);
  fclose(pfmed);
  qsort(leaves, NE, sizeof(struct edge), leaves_order);

  /* more initial values */
  total_vcross = NE;
  num_ecross = init_ecross = (float)NP * (float)(NP-1) / 2;
  for (i=0; i<NP; i++) {
    numv[i]=0;
    Pname[i]=i;
    Psize[i]=1;
  }

  init_reps2l(); /* initialize Urep, Vrep, UVleaves*/
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void init_reps2l()
{
  int u, v, tmp, n;
  struct vcross *p;

  for (n=0; n<NP; n++)
    Urep[n] = NULL;

  for (n=0; n<NE; n++) {
    u = leaves[n].u;
    v = leaves[n].v;

    /* try to balance the length of v-lists */
    if (numv[u] > numv[v]) {
      tmp = u;
      u = v;
      v = tmp;
    }

    /* enter the name (v) and the leaf (n) into Vrep and UVleaves */
    p = Vrep+n;
    p->vn = v;
    p->next = NULL;
    p->leaves = p->tail = UVleaves+n;

    p->leaves->leaf = n;
    p->leaves->next = NULL;

    /* put the new v-record at the beginning of the u-list */
    numv[u]++;
    p->next = Urep[u];
    Urep[u] = p;
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* get u1 and u2 to be merged.
   exchange u1 and u2 if Psize[u1] < Psize[u2].
   create a new list under u1 and delete u2, also update Pname and Psize.
   return a pointer to a list of crossing edges between u1 and u2. */
struct xrs *find_crossing(int u1, int u2)
{
  struct vcross *wksp, *p, *vloc;
  struct xrs *p2cross = NULL;
  int r, i, count;
  int u, vn, name;
  int indx, tmp;

  /* decrease total_vcross (will be increased below) */
  total_vcross -= (numv[u1]+numv[u2]);
  /* decrease number of crossing edges */
  delta_ecross = (float)Psize[u1] * (float)Psize[u2];
  num_ecross -= delta_ecross;
  /* modified delta_ecross: */
  if (Psize[u1] >= MIN_CLUSTER_SIZE && Psize[u2] >= MIN_CLUSTER_SIZE)
    delta_ecross2 = (float)Psize[u1] * (float)Psize[u2];
  else
    delta_ecross2 = 0.0;

  /* always make the smaller tree a child of the larger in Pname */
  if (Psize[u1] < Psize[u2]) {
    tmp = u1;
    u1 = u2;
    u2 = tmp;
  }
  Psize[u1] += Psize[u2];
  Psize[u2] = 0;

  /* create workspace */
  wksp = (struct vcross *)malloc((numv[u1]+numv[u2])*sizeof(struct vcross));
  if (!wksp) error("allocation error");

  /* correct the vn fields of each u-list and copy to wksp  */
  count = 0;
  for (i=0; i<2; i++) {
    if (i==0) u=u1; else u=u2;
    p = Urep[u];
    while (p != NULL) {
      /* find names of representatives */
      vn = Pname[p->vn];
      while (vn != Pname[vn])
	vn = Pname[vn];

      /* update Pname (may help in the future) */
      name = (vn==u2) ? u1 : vn;
      indx = p->vn;
      while (Pname[indx] != vn) {
	tmp = Pname[indx];
	Pname[indx] = name;
	indx = tmp;
      }

      /* copy item into wksp */
      /* a label u1 is changed to u2, in order to have the same name for 
	 crossing edges in both lists */
      name = (vn==u1) ? u2 : vn;
      wksp[count].vn = name;
      wksp[count].leaves = p->leaves;
      wksp[count].tail   = p->tail;
      wksp[count].next   = p; /* use the "next" field to store self address */
      count++;
      p = p->next;
    }
  }

  /* sort wksp, concatenate lists with the same name */
  qsort(wksp, count, sizeof(struct vcross), order_vcross);
  r=0;
  while (r < count) {
    if ((i=r+1)==count)
      break;
    while (wksp[i].vn == wksp[r].vn) {
      p = wksp[r].next; /* the real address in Vrep of item wksp[r] */
      p->tail->next = wksp[i].leaves;
      p->tail = wksp[i].tail;
      if (++i == count)
	break;
    }
    r=i;
  }

  /* build the list of u1, store a pointer to crossing edges */
  /* transfer records to other lists if they are shorter */
  Urep[u1] = NULL;
  numv[u1] = 0;		  
  r=0;
  while (r < count) {
    name = wksp[r].vn;
    vloc = wksp[r].next;

    if (name != u2) { /* the v-record is not the crossing one */

      if (numv[u1] <= numv[name]) {
	/* add the record into u1 list */ 
	vloc->next = Urep[u1];
	Urep[u1]   = vloc;
	
	Urep[u1]->vn = name;
	numv[u1]++;
	total_vcross++;
      } else {
	/* add the record into v-list */
	vloc->next = Urep[name];
	Urep[name] = vloc;

	Urep[name]->vn = u1;
	numv[name]++;
	total_vcross++;
      }

    } else { /* the cross u1-u2 */
      if (p2cross != NULL)
	error("p2cross is already assigned");
      p2cross = wksp[r].leaves;
    }

    if ((i=r+1)==count)
      break;
    while (wksp[i].vn == name) {
      if (++i == count)
	break;
    }
    r=i;
  }

  /* u2 is no longer a valid name */
  numv[u2] = 0;
  Urep[u2] = NULL;
  Pname[u2] = u1;

  free(wksp);
  return(p2cross);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int order_vcross(const void *x, const void *y)
{
  struct vcross *a, *b;

  a = (struct vcross *)x;
  b = (struct vcross *)y;

  /* sort by increasing order */
  if (a->vn < b->vn) return (-1);
  if (a->vn > b->vn) return ( 1);
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int order_partition(const void *x, const void *y)
{
  struct partition *a, *b;

  a = (struct partition *)x;
  b = (struct partition *)y;

  /* sort by decreasing order */
  if (a->height < b->height) return ( 1);
  if (a->height > b->height) return (-1);
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int leaves_order(const void *x, const void *y)
{
  struct edge *a, *b;

  a = (struct edge *)x;
  b = (struct edge *)y;

  /* sort by increasing order */
  if (a->weight < b->weight) return (-1);
  if (a->weight > b->weight) return ( 1);

  /* internal sorting according to u is used to make sorted output */
  if (a->u < b->u) return (-1);
  if (a->u > b->u) return ( 1);
  
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void print_vectors(FILE *pf, int *Labels, char *fdata, int shift)
{
  FILE *pfvec;
  int i,j;
  int tag, maxtag;
  int n, dim;
  float **vec;

  /* get vectorial data, # np dim, then float coordinates */
  if ((pfvec=fopen(fdata,"r"))==NULL)
    error("cannot open file specified by -v option");
  fscanf(pfvec,"%*s %d %d", &n, &dim);
  if (n != NP)
    error("files specified by -v and -w are not consistent");
  
  vec = (float **)malloc(NP*sizeof(float *));
  if (!vec) error("allocation error (vec)");
  for (i=0; i<NP; i++) {
    vec[i] = (float *)malloc(dim*sizeof(float));
    if (!vec[i]) error("allocation error (vec[i])");
  }

  for (i=0; i<NP; i++)
    for (j=0; j<dim; j++)
      fscanf(pfvec,"%f", &vec[i][j]);
  fclose(pfvec);


  /* find largest tag which is below NP (see correct_zeros for details) */
  maxtag = 0;
  if (shift) {
    for (i=0; i<NP; i++) {
      if (Labels[i] > maxtag && Labels[i] < NP)
	maxtag = Labels[i];
    }
  }
    
  /* print <label> <x> <y> .... */
  for (i=0; i<NP; i++) {
    if ((tag=Labels[i]) > NP)
      tag = tag-NP+maxtag;
    fprintf(pf, "%d", tag);
    for (j=0; j<dim; j++)
      fprintf(pf, " %f", vec[i][j]);
    fprintf(pf, "\n");
  }

  /* free vec */
  for (i=NP-1; i>=0; i--)
    free(vec[i]);
  free(vec);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void print_image(FILE *pf, int *Labels, char *fdata, int printype, int shift)
{
  FILE *pfimg;
  int nr,nc;
  int i,j;
  int tag, maxtag;

  if (printype == 2) {
    /* get image dimensions from header file */
    if ((pfimg=fopen(fdata,"r"))==NULL)
      error("cannot open file specified by -i option");
    fscanf(pfimg,"%*s %d %d", &nr, &nc);
    fclose(pfimg);
    if (nr*nc != NP)
      error("files specified by -i and -w are not consistent");
  } else { /* Par.printype == 3 */
    nr=1;
    nc=NP;
  }

  /* find largest tag which is below NP (see correct_zeros for details) */
  maxtag = 0;
  if (shift) {
    for (i=0; i<NP; i++) {
      if (Labels[i] > maxtag && Labels[i] < NP)
	maxtag = Labels[i];
    }
  }

  /* print image */
  for (i=0; i<nr; i++) {
    for (j=0; j<nc; j++) {
      if ((tag=Labels[i*nc+j]) > NP)
	tag = tag-NP+maxtag;
      fprintf(pf, "%d ", tag);
    }
    fprintf(pf, "\n");
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void print_indices(FILE *pf, int *Labels)
{
  int i, tag;

  /* use the array leaves as workspace to arrange the output */
  for (i=0; i<NP; i++) {
    leaves[i].weight = Labels[i];   /* cluster tag */
    leaves[i].u = i;                /* point id */
  }
  qsort(leaves, NP, sizeof(struct edge), leaves_order);


  /* print the indices of points in each cluster */
  i=0;
  while (i<NP) {
    fprintf(pf,"\n");
    tag = leaves[i].weight;
    if (tag==0)
      fprintf(pf,"BG> ");
    else
      fprintf(pf,"--> ");
    do {
      fprintf(pf,"%d ", leaves[i].u);
      if (++i==NP)
	break;
    } while (leaves[i].weight == tag);
    fprintf(pf,"\n");
  } 
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void command_line(int argc, char *argv[], struct params *Par)
{
  int i;
  char *data, c;

  /* initialize parameters to defaults */
  Par->fsim = Par->fmed = Par->fpeak = Par->fdata = NULL;
  Par->graph = 0;
  Par->correct0 = 0;
  Par->min_size = 2;
  Par->min_height = 0.01;
  Par->report_by = Par->report_val = Par->printype = 0;
  Par->quite = 0;

  /* set parameters according to command line */
  i=1;
  while (i<argc) {
    data = argv[i];

    /* expect to get an isolated letter with minus sign */
    c = data[0];
    if (c != '-') error("wrong usage");
    c = data[2];
    if (c != '\0') error("wrong usage");

    /* assign value */
    c = data[1];
    switch (c) {
    case 'w':
      if (++i == argc) error("wrong usage");
      Par->fsim = argv[i];
      break;
    case 'm':
      if (++i == argc) error("wrong usage");
      Par->fmed = argv[i];
      break;
    case 'p':
      if (++i == argc) error("wrong usage");
      Par->fpeak = argv[i];
      break;
    case 'v':
      if (++i == argc) error("wrong usage");
      Par->fdata = argv[i];
      if (Par->printype > 0) error("options v,i are mutually exclusive");
      Par->printype = 1;
      break;
    case 'i':
      if (Par->printype > 0) error("options v,i are mutually exclusive");

      if (++i == argc) /* i flag, no argument */
	Par->printype = 3;
      else
	if (argv[i][0] == '-') { /* i flag, no argument */
	  Par->printype = 3;
	  i--;
	} else {
	  Par->fdata = argv[i];
	  Par->printype = 2;
	}

      break;
    case 'g':
      Par->graph = 1;
      break;
    case 'l':
      if (++i == argc) error("wrong usage");
      Par->correct0 = atoi(argv[i]);
      if (Par->correct0 > 2) error("-l value either 0,1,2");
      break;
    case 'q':
      Par->quite = 1;
      break;
    case 's':
      if (++i == argc) error("wrong usage");
      Par->min_size = atoi(argv[i]);
      break;
    case 'h':
      if (++i == argc) error("wrong usage");
      Par->min_height = atof(argv[i]);
      break;
    case 'r':
      if (++i == argc) error("wrong usage");
      Par->report_val = atoi(argv[i]);
      if (Par->report_by > 0) error("options r,n,o are mutually exclusive");
      Par->report_by = 1;
      break;
    case 'n':
      if (++i == argc) error("wrong usage");
      Par->report_val = atoi(argv[i]);
      if (Par->report_by > 0) error("options r,n,o are mutually exclusive");
      Par->report_by = 2;
      break;
    case 'o':
      if (++i == argc) error("wrong usage");
      Par->report_val = atoi(argv[i]);
      if (Par->report_by > 0) error("options r,n,o are mutually exclusive");
      Par->report_by = 3;
      break;
    default:
      error("wrong usage");
      break;
    }

    i++;
  }

  /* check validity */
  if (Par->fsim == NULL) error("missing usage parameter: w");
  if (Par->fmed == NULL) error("missing usage parameter: m");
  if (!Par->graph && Par->fpeak == NULL) error("missing usage parameter: p");

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
